package config_test

import (
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"

	. "github.com/pivotal-cf-experimental/mysql-metrics/config"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("Config", func() {
	var (
		config                    *Config
		configFilepath            string
		tempDir                   string
		password                  string
		username                  string
		instanceId                string
		metricFrequency           int
		origin                    string
		emitLeaderFollowerMetrics bool
		heartbeatDatabase         string
		heartbeatTable            string
	)

	BeforeEach(func() {
		config = &Config{}
	})

	Describe("with a fully formed, correct yaml file", func() {
		BeforeEach(func() {
			username = "user"
			password = "secret"
			origin = "p-mysql"
			instanceId = "vm-123456"
			metricFrequency = 1
			emitLeaderFollowerMetrics = true
			heartbeatDatabase = "someDatabase"
			heartbeatTable = "someTable"
			tempDir, err := ioutil.TempDir("", "")
			Expect(err).NotTo(HaveOccurred())

			configFilepath = filepath.Join(tempDir, "metric-config.yml")
			configString := fmt.Sprintf(`{
				"instance_id":"%s",
				"username":"%s",
				"password":"%s",
				"metrics_frequency":%d,
				"origin":"%s",
				"emit_leader_follower_metrics":%t,
				"heartbeat_database":"%s",
				"heartbeat_table":"%s"
			}`, instanceId, username, password, metricFrequency, origin, emitLeaderFollowerMetrics, heartbeatDatabase, heartbeatTable)

			err = ioutil.WriteFile(configFilepath, []byte(configString), os.ModePerm)
			Expect(err).NotTo(HaveOccurred())
		})

		AfterEach(func() {
			err := os.RemoveAll(tempDir)
			Expect(err).NotTo(HaveOccurred())
		})

		It("reads the config file", func() {
			var err error
			err = LoadFromFile(configFilepath, config)

			Expect(err).NotTo(HaveOccurred())
			Expect(config).NotTo(BeNil())
			Expect(config.InstanceId).To(Equal(instanceId))
			Expect(config.Username).To(Equal(username))
			Expect(config.Password).To(Equal(password))
			Expect(config.MetricsFrequency).To(Equal(metricFrequency))
			Expect(config.Origin).To(Equal(origin))
			Expect(config.EmitLeaderFollowerMetrics).To(Equal(emitLeaderFollowerMetrics))
			Expect(config.HeartbeatDatabase).To(Equal(heartbeatDatabase))
			Expect(config.HeartbeatTable).To(Equal(heartbeatTable))
		})
	})

	Describe("when the yaml file is not fully formed", func() {
		BeforeEach(func() {
			configString := `"field1value1"}`
			tempDir, err := ioutil.TempDir("", "")
			Expect(err).NotTo(HaveOccurred())

			configFilepath = filepath.Join(tempDir, "metric-config.yml")

			err = ioutil.WriteFile(configFilepath, []byte(configString), os.ModePerm)
			Expect(err).NotTo(HaveOccurred())
		})

		AfterEach(func() {
			err := os.RemoveAll(tempDir)
			Expect(err).NotTo(HaveOccurred())
		})

		It("returns an error", func() {
			var err error
			err = LoadFromFile(configFilepath, config)

			Expect(err).To(HaveOccurred())
		})
	})

	Describe("when the yaml file does not exist", func() {
		It("returns an error", func() {
			var err error
			err = LoadFromFile("path/doesnot/exist", config)

			Expect(err).To(HaveOccurred())
		})
	})
})
