package database_client

//go:generate counterfeiter . DatabaseClient

import (
	"database/sql"
	"strings"

	configPackage "github.com/pivotal-cf-experimental/mysql-metrics/config"
)

type DbClient struct {
	connection *sql.DB
	config     *configPackage.Config
}

type DatabaseClient interface {
	ShowGlobalStatus() (map[string]string, error)
	ShowGlobalVariables() (map[string]string, error)
	ShowSlaveStatus() (map[string]string, error)
	SecondsSinceLeaderHeartbeat() (map[string]string, error)
}

func QuoteIdentifier(identifier string) string {
	return "`" + strings.Replace(identifier, "`", "``", -1) + "`"
}

func NewDatabaseClient(connection *sql.DB, config *configPackage.Config) *DbClient {
	return &DbClient{
		connection: connection,
		config:     config,
	}
}

func (dc *DbClient) ShowGlobalStatus() (map[string]string, error) {
	return dc.runKeyValueQuery("SHOW GLOBAL STATUS")
}

func (dc *DbClient) ShowGlobalVariables() (map[string]string, error) {
	return dc.runKeyValueQuery("SHOW GLOBAL VARIABLES")
}

func (dc *DbClient) ShowSlaveStatus() (map[string]string, error) {
	return dc.runSingleRowQuery("SHOW SLAVE STATUS", []interface{}{})
}

func (dc *DbClient) SecondsSinceLeaderHeartbeat() (map[string]string, error) {
	sql := "SELECT NOW() - timestamp AS seconds_since_leader_heartbeat FROM " +
		QuoteIdentifier(dc.config.HeartbeatDatabase) + "." + QuoteIdentifier(dc.config.HeartbeatTable)
	return dc.runSingleRowQuery(sql, []interface{}{})
}

func (dc *DbClient) runSingleRowQuery(query string, params []interface{}) (map[string]string, error) {
	rows, err := dc.connection.Query(query, params...)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	result := make(map[string]string)

	columns, err := rows.Columns()
	if err != nil {
		return nil, err
	}

	if rows.Next() {
		columnValues := make([]interface{}, len(columns))
		columnPointers := make([]interface{}, len(columns))
		for i := range columns {
			columnPointers[i] = &columnValues[i]
		}
		err := rows.Scan(columnPointers...)
		if err != nil {
			panic(err)
		}
		for index, col := range columns {
			value := columnValues[index]
			if value == nil {
				value = []uint8("NULL")
			}
			result[strings.ToLower(col)] = string(value.([]uint8))
		}
	}

	return result, nil
}

func (dc *DbClient) runKeyValueQuery(query string) (map[string]string, error) {
	rows, err := dc.connection.Query(query)
	if err != nil {
		return nil, err
	}
	defer rows.Close()

	globalStatusVariables := make(map[string]string)
	for rows.Next() {
		var key, value string
		rows.Scan(&key, &value)
		globalStatusVariables[strings.ToLower(key)] = value

	}

	return globalStatusVariables, nil
}
