package database_client_test

import (
	"database/sql"
	"errors"

	"github.com/DATA-DOG/go-sqlmock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	configPackage "github.com/pivotal-cf-experimental/mysql-metrics/config"
	"github.com/pivotal-cf-experimental/mysql-metrics/database_client"
)

var _ = Describe("DatabaseClient", func() {
	var (
		conn             *sql.DB
		mock             sqlmock.Sqlmock
		err              error
		dc               *database_client.DbClient
		config           *configPackage.Config
		hbDatabase       string
		hbTable          string
		hbTableQualified string
	)

	BeforeEach(func() {
		conn, mock, err = sqlmock.New()
		Expect(err).NotTo(HaveOccurred())

		hbDatabase = "someHbDatabase"
		hbTable = "someHbTable"
		hbTableQualified = database_client.QuoteIdentifier(hbDatabase) +
			"." + database_client.QuoteIdentifier(hbTable)

		config = &configPackage.Config{
			HeartbeatDatabase: hbDatabase,
			HeartbeatTable:    hbTable,
		}

		dc = database_client.NewDatabaseClient(conn, config)
	})

	Describe("ShowGlobalStatus", func() {
		It("returns a map of lowercased status variables and their values", func() {
			row := sqlmock.NewRows([]string{"variable_name", "value"}).
				AddRow("something", "doesntmatter").
				AddRow("Other Thing", "123.4")
			mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)

			status, err := dc.ShowGlobalStatus()
			Expect(err).NotTo(HaveOccurred())
			Expect(len(status)).To(Equal(2))
			Expect(status).To(Equal(map[string]string{
				"something":   "doesntmatter",
				"other thing": "123.4",
			}))
		})

		It("returns an error and no data when the query fails", func() {
			mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnError(errors.New("db unavailable"))
			_, err := dc.ShowGlobalStatus()
			Expect(err).To(MatchError("db unavailable"))
		})
	})

	Describe("ShowGlobalVariables", func() {
		It("returns a map of lowercased status variables and their values", func() {
			row := sqlmock.NewRows([]string{"variable_name", "value"}).
				AddRow("Max_connections", "12")
			mock.ExpectQuery(`SHOW GLOBAL VARIABLES`).WillReturnRows(row)

			vars, err := dc.ShowGlobalVariables()
			Expect(err).NotTo(HaveOccurred())
			Expect(vars).To(Equal(map[string]string{
				"max_connections": "12",
			}))
		})

		It("returns an error and no data when the query fails", func() {
			mock.ExpectQuery(`SHOW GLOBAL VARIABLES`).WillReturnError(errors.New("db unavailable"))
			_, err := dc.ShowGlobalVariables()
			Expect(err).To(MatchError("db unavailable"))
		})
	})

	Describe("ShowSlaveStatus", func() {
		It("returns an empty map when SHOW SLAVE STATUS returns an empty result (leader node)", func() {
			mock.ExpectQuery("SHOW SLAVE STATUS").WillReturnRows(sqlmock.NewRows([]string{}))

			vars, err := dc.ShowSlaveStatus()
			Expect(err).NotTo(HaveOccurred())
			Expect(vars).To(Equal(map[string]string{}))
		})

		It("returns a map of lowercased slave status variables and their values when SHOW SLAVE STATUS returns a non-empty result", func() {
			row := sqlmock.NewRows([]string{
				"Some_Slave_Status_Metric",
				"Slave_SQL_Running_State",
				"SQL_Lib_Not_Able_To_Parse",
			}).AddRow(
				[]uint8("123.4"),
				nil,
				[]uint8(nil),
			)
			mock.ExpectQuery("SHOW SLAVE STATUS").WillReturnRows(row)

			vars, err := dc.ShowSlaveStatus()
			Expect(err).NotTo(HaveOccurred())
			Expect(vars).To(Equal(map[string]string{
				"some_slave_status_metric":  "123.4",
				"slave_sql_running_state":   "NULL",
				"sql_lib_not_able_to_parse": "",
			}))
		})

		It("returns an error and no data when the query fails", func() {
			mock.ExpectQuery(`SHOW SLAVE STATUS`).WillReturnError(errors.New("db unavailable"))
			_, err := dc.ShowSlaveStatus()
			Expect(err).To(MatchError("db unavailable"))
		})
	})

	Describe("SecondsSinceLeaderHeartbeat", func() {
		It("returns a map containing seconds_since_leader_heartbeat", func() {
			row := sqlmock.NewRows([]string{
				"seconds_since_leader_heartbeat",
			}).AddRow(
				[]uint8("3"),
			)

			mock.ExpectQuery(
				`SELECT NOW\(\) - timestamp AS seconds_since_leader_heartbeat FROM ` +
					hbTableQualified).
				WillReturnRows(row)

			status, err := dc.SecondsSinceLeaderHeartbeat()
			Expect(err).NotTo(HaveOccurred())
			Expect(len(status)).To(Equal(1))
			Expect(status).To(Equal(map[string]string{
				"seconds_since_leader_heartbeat": "3",
			}))
		})

		It("returns an error and no data when the query fails", func() {
			mock.ExpectQuery(
				`SELECT NOW\(\) - timestamp AS seconds_since_leader_heartbeat FROM ` + hbTableQualified).
				WillReturnError(errors.New("db unavailable"))
			_, err := dc.SecondsSinceLeaderHeartbeat()
			Expect(err).To(MatchError("db unavailable"))
		})
	})

	Describe("QuoteIdentifier", func() {
		It("quotes identifier while escaping existing quotes", func() {
			Expect(database_client.QuoteIdentifier("foobar")).To(Equal("`foobar`"))
			Expect(database_client.QuoteIdentifier("bar`baz")).To(Equal("`bar``baz`"))
		})
	})
})
