package main

import (
	"database/sql"
	"flag"
	"fmt"
	"os"
	"time"

	"code.cloudfoundry.org/lager"
	"github.com/cloudfoundry/dropsonde"
	configNamespace "github.com/pivotal-cf-experimental/mysql-metrics/config"
	"github.com/pivotal-cf-experimental/mysql-metrics/database_client"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics_computer"
)

const (
	defaultConfigPath              = "/var/vcap/jobs/mysql-metrics/config/mysql-config.yml"
	defaultMetricMappingConfigPath = "/var/vcap/jobs/mysql-metrics/config/metric-mapping-config.yml"
)

type lagerLoggerWrapper struct {
	logger lager.Logger
}

func (d lagerLoggerWrapper) Debug(action string, message map[string]interface{}) {
	data := lager.Data{}

	for k, v := range message {
		data[k] = v
	}

	d.logger.Debug(action, data)
}

func (d lagerLoggerWrapper) Info(action string) {
	d.logger.Info(action)
}

func (d lagerLoggerWrapper) Error(action string, err error) {
	d.logger.Error(action, err)
}

var configFilepath = flag.String("c", defaultConfigPath, "location of config file")
var logFilepath = flag.String("l", "", "location of log file")
var mappingFilepath = flag.String("m", defaultMetricMappingConfigPath, "location of metric mapping config file")

func setupLogging(metricsLogger lager.Logger) {
	if *logFilepath != "" {
		file, err := os.OpenFile(*logFilepath, os.O_WRONLY|os.O_APPEND|os.O_CREATE, 0660)
		if err != nil {
			panic(fmt.Sprintf("Could not open log file: %s\n", err))
		}

		sink := lager.NewWriterSink(file, lager.DEBUG)
		metricsLogger.RegisterSink(sink)
	}
}

func main() {
	flag.Parse()

	metricsLogger := lager.NewLogger("MetricsLogger")
	sink := lager.NewWriterSink(os.Stdout, lager.DEBUG)
	metricsLogger.RegisterSink(sink)

	setupLogging(metricsLogger)

	config := &configNamespace.Config{}
	err := configNamespace.LoadFromFile(*configFilepath, config)

	if err != nil {
		metricsLogger.Error("config file is not formatted correctly", err)
		panic(err)
	}

	metricMappingConfig := &metrics.MetricMappingConfig{}
	err = configNamespace.LoadFromFile(*mappingFilepath, metricMappingConfig)

	if err != nil {
		metricsLogger.Error("metric mapping config file is not formatted correctly", err)
		panic(err)
	}

	err = dropsonde.Initialize("localhost:3457", config.Origin)
	if err != nil {
		metricsLogger.Error("dropsonde failed to initialize", err)
		panic(err)
	}

	conn := Connection(config)
	dbClient := database_client.NewDatabaseClient(conn, config)
	loggerWrapper := lagerLoggerWrapper{metricsLogger}
	client := metrics_computer.NewMetricsComputer(dbClient, loggerWrapper)
	metricsWriter := metrics.NewMetricWriter(new(metrics.DropsondeSender), loggerWrapper)
	emitLeaderFollowerMetrics := config.EmitLeaderFollowerMetrics

	process(client, metricsWriter, *metricMappingConfig, emitLeaderFollowerMetrics, metricsLogger)
	metricsInterval := time.Duration(config.MetricsFrequency) * time.Second
	for {
		select {
		case <-time.After(metricsInterval):
			process(client, metricsWriter, *metricMappingConfig, emitLeaderFollowerMetrics, metricsLogger)
		}
	}
}

func process(metricsComputer *metrics_computer.MetricsComputer,
	metricsWriter *metrics.MetricWriter,
	mappings metrics.MetricMappingConfig,
	emitLeaderFollowerMetrics bool,
	logger lager.Logger) {
	metrics, err := metricsComputer.QueryMetrics(mappings, emitLeaderFollowerMetrics)

	if err != nil {
		logger.Error("Could not get metrics:", err)
	}
	logger.Debug(fmt.Sprintf("Logging %d metrics", len(metrics)))
	for i := range metrics {
		metric := metrics[i]
		if metric.HasError {
			logger.Debug("Metric had error converting to a float", map[string]interface{}{"errorMetric": metric})
		} else {
			if err := metricsWriter.Write(metric); err != nil {
				logger.Error("Failed to send metric", err, map[string]interface{}{"errorMetric": metric})
			}
		}
	}
}

func Connection(config *configNamespace.Config) *sql.DB {
	conn, err := sql.Open("mysql", fmt.Sprintf("%s:%s@tcp(%s:%d)/", config.Username, config.Password, "localhost", 3306))
	if err != nil {
		panic(fmt.Sprintf("Database configuration problem: %v", err))
	}

	return conn
}
