package metrics_computer

import (
	"errors"
	"strconv"

	"strings"

	_ "github.com/go-sql-driver/mysql"
	"github.com/pivotal-cf-experimental/mysql-metrics/database_client"
	. "github.com/pivotal-cf-experimental/mysql-metrics/logger"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
)

type MetricsComputer struct {
	client database_client.DatabaseClient
	logger MetricsLogger
}

func NewMetricsComputer(client database_client.DatabaseClient, logger MetricsLogger) *MetricsComputer {
	return &MetricsComputer{
		client: client,
		logger: logger,
	}
}

func (mc *MetricsComputer) QueryMetrics(mappings metrics.MetricMappingConfig, emitLeaderFollowerMetrics bool) ([]*metrics.Metric, error) {
	mysqlUnavailable := []*metrics.Metric{mc.availableMetric(false, mappings.GetInternalMetricMappings())}

	globalStatus, err := mc.client.ShowGlobalStatus()
	if err != nil {
		return mysqlUnavailable, err
	}

	globalVariables, err := mc.client.ShowGlobalVariables()
	if err != nil {
		return mysqlUnavailable, err
	}

	var gatheredMetrics []*metrics.Metric
	gatheredMetrics = append(gatheredMetrics, mc.addMetrics(globalStatus, mappings.GetMysqlMetricMappings())...)
	gatheredMetrics = append(gatheredMetrics, mc.addMetrics(globalVariables, mappings.GetMysqlMetricMappings())...)

	if emitLeaderFollowerMetrics {
		leaderFollowerMetrics, err := mc.getLeaderFollowerMetrics(mappings.GetLeaderFollowerMetricMappings())
		if err != nil {
			return mysqlUnavailable, err
		}

		gatheredMetrics = append(gatheredMetrics, leaderFollowerMetrics...)
	}

	gatheredMetrics = append(gatheredMetrics, mc.availableMetric(true, mappings.GetInternalMetricMappings()))

	return gatheredMetrics, nil
}

func (mc *MetricsComputer) getLeaderFollowerMetrics(leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping) ([]*metrics.Metric, error) {
	var gatheredMetrics []*metrics.Metric

	slaveStatus, err := mc.client.ShowSlaveStatus()
	if err != nil {
		return nil, err
	}

	isFollowerMetric := mc.getIsFollowerMetric(slaveStatus, leaderFollowerMetricMappings)
	gatheredMetrics = append(gatheredMetrics, isFollowerMetric)

	if isFollowerMetric.Value == 1.0 {
		secondsSinceLeaderHeartbeat, err := mc.client.SecondsSinceLeaderHeartbeat()
		if err == nil {
			secondsSinceLeaderHeartbeatMetric := mc.getFollowerMetrics(secondsSinceLeaderHeartbeat, leaderFollowerMetricMappings)
			gatheredMetrics = append(gatheredMetrics, secondsSinceLeaderHeartbeatMetric...)
		} else {
			mc.logger.Error("Error computing seconds since leader heartbeat", err)
		}

		slaveStatusMetrics := mc.getFollowerMetrics(slaveStatus, leaderFollowerMetricMappings)
		gatheredMetrics = append(gatheredMetrics, slaveStatusMetrics...)
	}

	return gatheredMetrics, nil
}

func (mc *MetricsComputer) getFollowerMetrics(slaveStatus map[string]string, leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping) []*metrics.Metric {
	var gatheredMetrics []*metrics.Metric

	for metricName, mapping := range leaderFollowerMetricMappings {
		value, found := slaveStatus[metricName]
		if found {
			metric := metrics.Metric{
				Key:      mapping.Key,
				Unit:     mapping.Unit,
				RawValue: value,
			}
			metric = mc.metricWithValue(&metric)
			gatheredMetrics = append(gatheredMetrics, &metric)
		}
	}
	return gatheredMetrics
}

func (mc *MetricsComputer) addMetrics(dbVars map[string]string, metricMappings map[string]metrics.MysqlMetricMapping) []*metrics.Metric {
	var gatheredMetrics []*metrics.Metric

	for metricName, mapping := range metricMappings {
		value, found := dbVars[metricName]
		if found {
			metric := metrics.Metric{
				Key:      mapping.Key,
				Unit:     mapping.Unit,
				RawValue: value,
			}
			metric = mc.metricWithValue(&metric)
			gatheredMetrics = append(gatheredMetrics, &metric)
		}
	}
	return gatheredMetrics
}

func (mc *MetricsComputer) getIsFollowerMetric(dbVars map[string]string, leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping) *metrics.Metric {
	isFollower := 0.0
	if len(dbVars) > 0 {
		isFollower = 1.0
	}
	key := leaderFollowerMetricMappings["is_follower"].Key
	return &metrics.Metric{Key: key, Value: isFollower, Unit: "bool"}
}

func (mc *MetricsComputer) metricWithValue(metric *metrics.Metric) metrics.Metric {
	floatValue, err := mc.parseFloat(metric.RawValue)
	if err != nil {
		floatValue, err = mc.parseStringValue(metric.RawValue)
	}
	if err != nil {
		floatValue, err = mc.parseClusterStatus(metric.RawValue)
	}

	return metrics.Metric{
		Key:      metric.Key,
		Unit:     metric.Unit,
		RawValue: metric.RawValue,
		Value:    floatValue,
		HasError: err != nil,
	}
}

func (mc *MetricsComputer) availableMetric(isAvailable bool, internalMetricMappings map[string]metrics.InternalMetricMapping) *metrics.Metric {
	availableAsFloat := 0.0
	if isAvailable {
		availableAsFloat = 1.0
	}
	key := internalMetricMappings["available"].Key

	return &metrics.Metric{Key: key, Value: availableAsFloat, Unit: "bool"}
}

func (mc *MetricsComputer) parseFloat(value string) (float64, error) {
	return strconv.ParseFloat(value, 64)
}

func (mc *MetricsComputer) parseStringValue(value string) (float64, error) {
	var floatValue float64
	var err error

	switch strings.ToLower(value) {
	case "on":
		floatValue = 1.0
	case "off":
		floatValue = 0.0
	case "yes":
		floatValue = 1.0
	case "no":
		floatValue = 0.0
	case "null":
		floatValue = -1.0
	default:
		err = errors.New("could not convert wsrep_ready")
	}

	return floatValue, err
}

func (mc *MetricsComputer) parseClusterStatus(value string) (float64, error) {
	var floatValue float64
	var err error

	switch strings.ToLower(value) {
	case "primary":
		floatValue = 1
	case "non-primary":
		floatValue = 0
	case "disconnected":
		floatValue = -1
	default:
		err = errors.New("could not convert wsrep_cluster_status")
	}

	return floatValue, err
}
