package metrics_computer_test

import (
	"errors"

	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics_computer"

	"strings"

	"fmt"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf-experimental/mysql-metrics/database_client/database_clientfakes"
	"github.com/pivotal-cf-experimental/mysql-metrics/logger/loggerfakes"
)

var _ = Describe("MetricsComputer", func() {
	var (
		databaseClient               *database_clientfakes.FakeDatabaseClient
		fakeLogger                   *loggerfakes.FakeMetricsLogger
		metricsComputer              *metrics_computer.MetricsComputer
		mysqlMetricMappings          map[string]metrics.MysqlMetricMapping
		internalMetricMappings       map[string]metrics.InternalMetricMapping
		leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping
		metricMappingConfig          metrics.MetricMappingConfig
		emitLeaderFollowerMetrics    bool
	)

	findMetricByKey := func(metrics []*metrics.Metric, key string) *metrics.Metric {
		for i := range metrics {
			aMetric := metrics[i]
			if strings.Contains(aMetric.Key, key) {
				return aMetric
			}
		}
		return nil
	}

	BeforeEach(func() {
		mysqlMetricMappings = map[string]metrics.MysqlMetricMapping{
			"questions":                     {Key: "/p.mysql/global_statuses/questions", Unit: "metric"},
			"innodb_buffer_pool_pages_free": {Key: "/p.mysql/innodb/buffer_pool_pages_free", Unit: "byte"},
			"innodb_data_read":              {Key: "/p.mysql/innodb/data_read", Unit: "byte"},
			"wsrep_cluster_status":          {Key: "/p.mysql/galera/wsrep_cluster_status", Unit: "number"},
			"wsrep_ready":                   {Key: "/p.mysql/galera/wsrep_ready", Unit: "bool"},
			"max_connections":               {Key: "/p.mysql/variables/max_connections", Unit: "integer"},
		}
		internalMetricMappings = map[string]metrics.InternalMetricMapping{
			"available": {Key: "/p.mysql/available"},
		}
		leaderFollowerMetricMappings = map[string]metrics.LeaderFollowerMetricMapping{
			"is_follower":                                    {Key: "/p.mysql/follower/is_follower", Unit: "bool"},
			"seconds_behind_master":                          {Key: "/p.mysql/follower/seconds_behind_master", Unit: "number"},
			"slave_sql_running":                              {Key: "/p.mysql/follower/slave_sql_running", Unit: "bool"},
			"some_random_invalid_key_should_not_get_emitted": {Key: "/p.mysql/follower/some_random_invalid_key_should_not_get_emitted", Unit: "number"},
			"seconds_since_leader_heartbeat":                 {Key: "/p.mysql/follower/seconds_since_leader_heartbeat", Unit: "number"},
		}

		metricMappingConfig = metrics.MetricMappingConfig{
			MysqlMetricMappings:          mysqlMetricMappings,
			InternalMetricMappings:       internalMetricMappings,
			LeaderFollowerMetricMappings: leaderFollowerMetricMappings,
		}

		databaseClient = new(database_clientfakes.FakeDatabaseClient)
		fakeLogger = new(loggerfakes.FakeMetricsLogger)
		metricsComputer = metrics_computer.NewMetricsComputer(databaseClient, fakeLogger)
	})

	Describe("QueryMetrics()", func() {
		Describe("when ShowGlobalStatus() returns an error", func() {
			BeforeEach(func() {
				emitLeaderFollowerMetrics = false
				databaseClient.ShowGlobalStatusReturns(nil, errors.New("query error"))
			})

			It("returns an error and database is down metric", func() {
				resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
				Expect(err).To(HaveOccurred())
				Expect(err).To(MatchError("query error"))
				Expect(resultMetrics).NotTo(BeNil())
				Expect(len(resultMetrics)).To(Equal(1))

				metric := resultMetrics[0]
				Expect(metric.Key).To(Equal("/p.mysql/available"))
				Expect(metric.Value).To(Equal(0.0))
				Expect(metric.HasError).To(BeFalse())
			})
		})

		Describe("when ShowGlobalStatus() returns non-float data", func() {
			BeforeEach(func() {
				emitLeaderFollowerMetrics = false
				databaseClient.ShowGlobalStatusReturns(map[string]string{
					"questions":                     "Nope",
					"innodb_buffer_pool_pages_free": "0",
				}, nil)

				databaseClient.ShowGlobalVariablesReturns(map[string]string{}, nil)
			})

			It("returns the metrics corresponding to the global status, with the HasError field of the malformed metric set to true", func() {
				resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())
				Expect(len(resultMetrics)).To(Equal(3))

				metric := findMetricByKey(resultMetrics, "global_statuses")
				Expect(metric.Key).To(Equal("/p.mysql/global_statuses/questions"))
				Expect(metric.RawValue).To(Equal("Nope"))
				Expect(metric.HasError).To(BeTrue())

				metric = findMetricByKey(resultMetrics, "buffer_pool_pages_free")
				Expect(metric.Key).To(Equal("/p.mysql/innodb/buffer_pool_pages_free"))
				Expect(metric.HasError).To(BeFalse())
			})
		})

		Describe("when ShowGlobalStatus() and ShowGlobalVariables() return expected data", func() {
			var (
				numQuestions      float64
				dataRead          float64
				maxConnections    float64
				clusterStatus     string
				galeraReadyStatus string
			)

			BeforeEach(func() {
				numQuestions = float64(125146.0)
				dataRead = float64(24121.0)
				clusterStatus = "Primary" // this might also be PRIMARY sometimes?
				galeraReadyStatus = "ON"
				maxConnections = 12
				emitLeaderFollowerMetrics = false
			})

			queriesReturnDefaultValues := func() {
				databaseClient.ShowGlobalStatusReturns(map[string]string{
					"something":            "doesntmatter",
					"questions":            fmt.Sprintf("%f", numQuestions),
					"innodb_data_read":     fmt.Sprintf("%f", dataRead),
					"wsrep_cluster_status": clusterStatus,
					"wsrep_ready":          galeraReadyStatus,
					"other thing":          "Still doesn't matter",
				}, nil)

				databaseClient.ShowGlobalVariablesReturns(map[string]string{
					"max_connections": fmt.Sprintf("%f", maxConnections),
				}, nil)
			}

			It("returns an array of metric objects", func() {
				queriesReturnDefaultValues()
				resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)

				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())
				Expect(len(resultMetrics)).To(Equal(6))

				aResultMetric := findMetricByKey(resultMetrics, "global_statuses")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/global_statuses/questions"))
				Expect(aResultMetric.Value).To(Equal(numQuestions))
				Expect(aResultMetric.Unit).To(Equal("metric"))

				aResultMetric = findMetricByKey(resultMetrics, "data_read")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/innodb/data_read"))
				Expect(aResultMetric.Value).To(Equal(dataRead))
				Expect(aResultMetric.Unit).To(Equal("byte"))

				aResultMetric = findMetricByKey(resultMetrics, "max_connections")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/variables/max_connections"))
				Expect(aResultMetric.Value).To(Equal(maxConnections))
				Expect(aResultMetric.Unit).To(Equal("integer"))

				aResultMetric = findMetricByKey(resultMetrics, "/p.mysql/available")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/available"))
				Expect(aResultMetric.Value).To(Equal(1.0))
				Expect(aResultMetric.Unit).To(Equal("bool"))
			})

			Describe("when emitLeaderFollowerMetrics is true", func() {
				BeforeEach(func() {
					emitLeaderFollowerMetrics = true
				})

				Describe("when ShowSlaveStatus() returns a query error", func() {
					BeforeEach(func() {
						queriesReturnDefaultValues()
						databaseClient.ShowSlaveStatusReturns(nil, errors.New("query error"))
					})

					It("emits only the available metric as false, and returns an error", func() {
						resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
						Expect(err).To(HaveOccurred())
						Expect(err).To(MatchError("query error"))
						Expect(resultMetrics).NotTo(BeNil())
						Expect(len(resultMetrics)).To(Equal(1))

						metric := findMetricByKey(resultMetrics, "/p.mysql/available")
						Expect(metric).NotTo(BeNil())
						Expect(metric.Value).To(Equal(0.0))
						Expect(metric.HasError).To(BeFalse())
					})
				})

				Describe("when ShowSlaveStatus() returns an empty result (leader node)", func() {
					BeforeEach(func() {
						queriesReturnDefaultValues()
						databaseClient.ShowSlaveStatusReturns(map[string]string{}, nil)
					})

					It("returns is_follower metric as 0", func() {
						resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
						Expect(err).ToNot(HaveOccurred())

						metric := findMetricByKey(resultMetrics, "/p.mysql/follower/is_follower")
						Expect(metric).NotTo(BeNil())
						Expect(metric.Value).To(Equal(0.0))
						Expect(metric.Unit).To(Equal("bool"))
					})
				})

				Describe("when ShowSlaveStatus() returns a non-empty result", func() {
					BeforeEach(func() {
						queriesReturnDefaultValues()
						databaseClient.ShowSlaveStatusReturns(map[string]string{
							"some-key": "is-not-empty",
						}, nil)
					})

					Describe("the is_follower metric", func() {
						It("is emitted with value 1.0", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/is_follower")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(1.0))
							Expect(metric.Unit).To(Equal("bool"))
						})
					})

					Describe("when seconds_behind_master is a number", func() {
						BeforeEach(func() {
							databaseClient.ShowSlaveStatusReturns(map[string]string{
								"seconds_behind_master": "2",
							}, nil)
						})

						It("emits the seconds_behind_master metric with value equal to the number", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_behind_master")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(2.0))
							Expect(metric.Unit).To(Equal("number"))
						})
					})

					Describe("when seconds_behind_master is NULL", func() {
						BeforeEach(func() {
							databaseClient.ShowSlaveStatusReturns(map[string]string{
								"seconds_behind_master": "NULL",
							}, nil)
						})

						It("emits the seconds_behind_master metric with value equal to -1 (TBD)", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_behind_master")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(-1.0))
							Expect(metric.Unit).To(Equal("number"))
						})
					})

					Describe("when Slave_SQL_Running is string 'No'", func() {
						BeforeEach(func() {
							databaseClient.ShowSlaveStatusReturns(map[string]string{
								"slave_sql_running": "No",
							}, nil)
						})

						It("emits the slave_sql_running metric with value equal to 0", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/slave_sql_running")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(0.0))
							Expect(metric.Unit).To(Equal("bool"))
						})
					})

					Describe("when Slave_SQL_Running is string Yes", func() {
						BeforeEach(func() {
							databaseClient.ShowSlaveStatusReturns(map[string]string{
								"slave_sql_running": "Yes",
							}, nil)
						})

						It("emits the slave_sql_running metric with value equal to 1", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/slave_sql_running")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(1.0))
							Expect(metric.Unit).To(Equal("bool"))
						})
					})

					Describe("when SecondsSinceLeaderHeartbeat returns a number", func() {
						BeforeEach(func() {
							databaseClient.SecondsSinceLeaderHeartbeatReturns(map[string]string{
								"seconds_since_leader_heartbeat": "3",
							}, nil)
						})

						It("emits the seconds_since_leader_heartbeat metric with value equal to the number", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_since_leader_heartbeat")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(3.0))
							Expect(metric.Unit).To(Equal("number"))
						})

						It("does not log an error", func() {
							metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)

							Expect(fakeLogger.ErrorCallCount()).To(Equal(0))
						})
					})

					Describe("when SecondsSinceLeaderHeartbeat returns an error", func() {
						BeforeEach(func() {
							databaseClient.SecondsSinceLeaderHeartbeatReturns(nil, errors.New("query error"))
						})

						It("does not emit the seconds_since_leader_heartbeat metric", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_since_leader_heartbeat")
							Expect(metric).To(BeNil())
						})

						It("logs the error so we can investigate why the heartbeat is busted", func() {
							metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)

							Expect(fakeLogger.ErrorCallCount()).To(Equal(1))
							errorString, errorErr := fakeLogger.ErrorArgsForCall(0)
							Expect(errorString).To(Equal("Error computing seconds since leader heartbeat"))
							Expect(errorErr).To(HaveOccurred())
							Expect(errorErr.Error()).To(Equal("query error"))
						})
					})

					Describe("when random/invalid key added in config", func() {
						BeforeEach(func() {
							databaseClient.ShowSlaveStatusReturns(map[string]string{
								"seconds_behind_master": "some-value",
							}, nil)
						})

						It("should not emit invalid key mapping from config", func() {
							resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/some_random_invalid_key_should_not_get_emitted")
							Expect(metric).To(BeNil())
						})
					})
				})
			})

			Describe("when emitLeaderFollowerMetrics is false", func() {
				BeforeEach(func() {
					emitLeaderFollowerMetrics = false
					queriesReturnDefaultValues()
				})

				It("should not emit is_follower metric", func() {
					resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					Expect(err).ToNot(HaveOccurred())
					for _, metric := range resultMetrics {
						Expect(metric.Key).ToNot(Equal("/p.mysql/follower/is_follower"))
					}
				})
			})

			Describe("when the data needs a special mapping", func() {
				BeforeEach(func() {
					queriesReturnDefaultValues()
				})

				It("should map cluster status to a number", func() {
					resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)

					Expect(err).NotTo(HaveOccurred())
					Expect(resultMetrics).NotTo(BeNil())

					metric := findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Primary maps to 1
					Expect(metric.Value).To(Equal(1.0))
					Expect(metric.Unit).To(Equal("number"))

					databaseClient.ShowGlobalStatusReturns(map[string]string{
						"wsrep_cluster_status": "Non-Primary",
					}, nil)

					resultMetrics, err = metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Non-Primary maps to 0
					Expect(metric.Value).To(Equal(0.0))

					databaseClient.ShowGlobalStatusReturns(map[string]string{
						"wsrep_cluster_status": "Disconnected",
					}, nil)

					resultMetrics, err = metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Disconnected maps to -1
					Expect(metric.Value).To(Equal(-1.0))

					databaseClient.ShowGlobalStatusReturns(map[string]string{
						"wsrep_cluster_status": "DISCONNECTED",
					}, nil)

					resultMetrics, err = metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Disconnected maps to -1
					Expect(metric.Value).To(Equal(-1.0))
					Expect(metric.HasError).To(BeFalse())
				})

				It("should map wsrep_ready to a 1 or 0", func() {
					resultMetrics, err := metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					Expect(err).NotTo(HaveOccurred())

					metric := findMetricByKey(resultMetrics, "wsrep_ready")
					Expect(metric).NotTo(BeNil())
					// ON to 1
					Expect(metric.Value).To(Equal(1.0))
					Expect(metric.Unit).To(Equal("bool"))

					databaseClient.ShowGlobalStatusReturns(map[string]string{
						"wsrep_ready": "OFF",
					}, nil)

					resultMetrics, err = metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_ready")
					// OFF to 0
					Expect(metric.Value).To(Equal(0.0))

					databaseClient.ShowGlobalStatusReturns(map[string]string{
						"wsrep_ready": "FoobarValue",
					}, nil)

					resultMetrics, err = metricsComputer.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_ready")
					Expect(resultMetrics).NotTo(BeNil())
					metric = findMetricByKey(resultMetrics, "wsrep_ready")
					Expect(metric).NotTo(BeNil())
					Expect(metric.RawValue).To(Equal("FoobarValue"))
					Expect(metric.HasError).To(BeTrue())
				})
			})
		})
	})
})
